"""
Bitbucket OAuth2 and OAuth1 backends, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/bitbucket.html
"""
from ..exceptions import AuthForbidden
from .oauth import BaseOAuth1, BaseOAuth2


class BitbucketOAuthBase(object):
    ID_KEY = 'uuid'

    def get_user_id(self, details, response):
        id_key = self.ID_KEY
        if self.setting('USERNAME_AS_ID', False):
            id_key = 'username'
        return response.get(id_key)

    def get_user_details(self, response):
        """Return user details from Bitbucket account"""
        fullname, first_name, last_name = self.get_user_names(
            response['display_name']
        )

        return {'username': response.get('username', ''),
                'email': response.get('email', ''),
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Return user data provided"""
        emails = self._get_emails(access_token)
        email = None

        for address in reversed(emails['values']):
            email = address['email']
            if address['is_primary']:
                break

        if self.setting('VERIFIED_EMAILS_ONLY', False) and \
            not address['is_confirmed']:
            raise AuthForbidden(
                self,
                'Bitbucket account has no verified email'
            )

        user = self._get_user(access_token)
        if email:
            user['email'] = email
        return user

    def _get_user(self, access_token=None):
        raise NotImplementedError('Implement in subclass')

    def _get_emails(self, access_token=None):
        raise NotImplementedError('Implement in subclass')


class BitbucketOAuth2(BitbucketOAuthBase, BaseOAuth2):
    name = 'bitbucket-oauth2'
    SCOPE_SEPARATOR = ' '
    AUTHORIZATION_URL = 'https://bitbucket.org/site/oauth2/authorize'
    ACCESS_TOKEN_URL = 'https://bitbucket.org/site/oauth2/access_token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ('scopes', 'scopes'),
        ('expires_in', 'expires'),
        ('token_type', 'token_type'),
        ('refresh_token', 'refresh_token')
    ]

    def auth_complete_credentials(self):
        return self.get_key_and_secret()

    def _get_user(self, access_token=None):
        return self.get_json('https://api.bitbucket.org/2.0/user',
                             params={'access_token': access_token})

    def _get_emails(self, access_token=None):
        return self.get_json('https://api.bitbucket.org/2.0/user/emails',
                             params={'access_token': access_token})

    def refresh_token(self, *args, **kwargs):
        raise NotImplementedError('Refresh tokens for Bitbucket have '
                                  'not been implemented')


class BitbucketOAuth(BitbucketOAuthBase, BaseOAuth1):
    """Bitbucket OAuth authentication backend"""
    name = 'bitbucket'
    AUTHORIZATION_URL = 'https://bitbucket.org/api/1.0/oauth/authenticate'
    REQUEST_TOKEN_URL = 'https://bitbucket.org/api/1.0/oauth/request_token'
    ACCESS_TOKEN_URL = 'https://bitbucket.org/api/1.0/oauth/access_token'

    def oauth_auth(self, *args, **kwargs):
        return super(BitbucketOAuth, self).oauth_auth(*args, **kwargs)

    def _get_user(self, access_token=None):
        return self.get_json('https://api.bitbucket.org/2.0/user',
                             auth=self.oauth_auth(access_token))

    def _get_emails(self, access_token=None):
        return self.get_json('https://api.bitbucket.org/2.0/user/emails',
                             auth=self.oauth_auth(access_token))



"""
Deezer backend, docs at:
    https://developers.deezer.com/api/oauth
    https://developers.deezer.com/api/permissions
"""
from six.moves.urllib.parse import parse_qsl

from .oauth import BaseOAuth2


class DeezerOAuth2(BaseOAuth2):
    """Deezer OAuth2 authentication backend"""
    name = 'deezer'
    ID_KEY = 'name'
    AUTHORIZATION_URL = 'https://connect.deezer.com/oauth/auth.php'
    ACCESS_TOKEN_URL = 'https://connect.deezer.com/oauth/access_token.php'
    ACCESS_TOKEN_METHOD = 'POST'
    SCOPE_SEPARATOR = ','
    REDIRECT_STATE = False

    def auth_complete_params(self, state=None):
        client_id, client_secret = self.get_key_and_secret()
        return {
            'app_id': client_id,
            'secret': client_secret,
            'code': self.data.get('code')
        }

    def request_access_token(self, *args, **kwargs):
        response = self.request(*args, **kwargs)
        return dict(parse_qsl(response.text))

    def get_user_details(self, response):
        """Return user details from Deezer account"""
        fullname, first_name, last_name = self.get_user_names(
            first_name=response.get('firstname'),
            last_name=response.get('lastname')
        )
        return {'username': response.get('name'),
                'email': response.get('email'),
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json('http://api.deezer.com/user/me', params={
            'access_token': access_token
        })



"""
Fedora OpenId backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/fedora.html
"""
from .open_id import OpenIdAuth


class FedoraOpenId(OpenIdAuth):
    name = 'fedora'
    URL = 'https://id.fedoraproject.org'
    USERNAME_KEY = 'nickname'



"""
Meetup OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/meetup.html
"""
from .oauth import BaseOAuth2


class MeetupOAuth2(BaseOAuth2):
    """Meetup OAuth2 authentication backend"""
    name = 'meetup'
    AUTHORIZATION_URL = 'https://secure.meetup.com/oauth2/authorize'
    ACCESS_TOKEN_URL = 'https://secure.meetup.com/oauth2/access'
    ACCESS_TOKEN_METHOD = 'POST'
    DEFAULT_SCOPE = ['basic']
    SCOPE_SEPARATOR = ','
    REDIRECT_STATE = False
    STATE_PARAMETER = 'state'

    def get_user_details(self, response):
        """Return user details from Meetup account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get('name')
        )

        return {'username': response.get('username'),
                'email': response.get('email') or '',
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json('https://api.meetup.com/2/member/self',
                             params={'access_token': access_token})



"""
Docker Hub OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/docker.html
"""
from .oauth import BaseOAuth2


class DockerOAuth2(BaseOAuth2):
    name = 'docker'
    ID_KEY = 'user_id'
    AUTHORIZATION_URL = 'https://hub.docker.com/api/v1.1/o/authorize/'
    ACCESS_TOKEN_URL = 'https://hub.docker.com/api/v1.1/o/token/'
    REFRESH_TOKEN_URL = 'https://hub.docker.com/api/v1.1/o/token/'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ('refresh_token', 'refresh_token', True),
        ('user_id', 'user_id'),
        ('email', 'email'),
        ('full_name', 'fullname'),
        ('location', 'location'),
        ('url', 'url'),
        ('company', 'company'),
        ('gravatar_email', 'gravatar_email'),
    ]

    def get_user_details(self, response):
        """Return user details from Docker Hub account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get('full_name') or response.get('username') or ''
        )
        return {
            'username': response.get('username'),
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name,
            'email': response.get('email', '')
        }

    def user_data(self, access_token, *args, **kwargs):
        """Grab user profile information from Docker Hub."""
        username = kwargs['response']['username']
        return self.get_json(
            'https://hub.docker.com/api/v1.1/users/%s/' % username,
            headers={'Authorization': 'Bearer %s' % access_token}
        )



"""
Foursquare OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/foursquare.html
"""
from .oauth import BaseOAuth2


class FoursquareOAuth2(BaseOAuth2):
    name = 'foursquare'
    AUTHORIZATION_URL = 'https://foursquare.com/oauth2/authenticate'
    ACCESS_TOKEN_URL = 'https://foursquare.com/oauth2/access_token'
    ACCESS_TOKEN_METHOD = 'POST'
    API_VERSION = '20140128'

    def get_user_id(self, details, response):
        return response['response']['user']['id']

    def get_user_details(self, response):
        """Return user details from Foursquare account"""
        info = response['response']['user']
        email = info['contact']['email']
        fullname, first_name, last_name = self.get_user_names(
            first_name=info.get('firstName', ''),
            last_name=info.get('lastName', '')
        )
        return {'username': first_name + ' ' + last_name,
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name,
                'email': email}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json('https://api.foursquare.com/v2/users/self',
                             params={'oauth_token': access_token,
                                     'v': self.API_VERSION})



import json

from .oauth import OAuth2Test


class PinterestOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.pinterest.PinterestOAuth2'
    user_data_url = 'https://api.pinterest.com/v1/me/'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'id': '4788400174839062',
        'first_name': 'Foo',
        'last_name': 'Bar',
        'username': 'foobar',
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()


class PinterestOAuth2BrokenServerResponseTest(OAuth2Test):
    backend_path = 'social_core.backends.pinterest.PinterestOAuth2'
    user_data_url = 'https://api.pinterest.com/v1/me/'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'data': {
            'id': '4788400174839062',
            'first_name': 'Foo',
            'last_name': 'Bar',
            'url': 'https://www.pinterest.com/foobar/',
        }
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Upwork OAuth1 backend
"""
from .oauth import BaseOAuth1


class UpworkOAuth(BaseOAuth1):
    """Upwork OAuth authentication backend"""
    name = 'upwork'
    ID_KEY = 'id'
    AUTHORIZATION_URL = 'https://www.upwork.com/services/api/auth'
    REQUEST_TOKEN_URL = \
        'https://www.upwork.com/api/auth/v1/oauth/token/request'
    REQUEST_TOKEN_METHOD = 'POST'
    ACCESS_TOKEN_URL = 'https://www.upwork.com/api/auth/v1/oauth/token/access'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_URI_PARAMETER_NAME = 'oauth_callback'

    def get_user_details(self, response):
        """Return user details from Upwork account"""
        info = response.get('info', {})
        auth_user = response.get('auth_user', {})
        first_name = auth_user.get('first_name')
        last_name = auth_user.get('last_name')
        fullname = '{} {}'.format(first_name, last_name)
        profile_url = info.get('profile_url', '')
        username = profile_url.rsplit('/')[-1].replace('~', '')
        return {
            'username': username,
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            'https://www.upwork.com/api/auth/v1/info.json',
            auth=self.oauth_auth(access_token)
        )



from ..exceptions import InvalidEmail
from .partial import partial


@partial
def mail_validation(backend, details, is_new=False, *args, **kwargs):
    requires_validation = backend.REQUIRES_EMAIL_VALIDATION or \
                          backend.setting('FORCE_EMAIL_VALIDATION', False)
    send_validation = details.get('email') and \
                      (is_new or backend.setting('PASSWORDLESS', False))
    if requires_validation and send_validation:
        data = backend.strategy.request_data()
        if 'verification_code' in data:
            backend.strategy.session_pop('email_validation_address')
            if not backend.strategy.validate_email(details['email'],
                                                   data['verification_code']):
                raise InvalidEmail(backend)
        else:
            current_partial = kwargs.get('current_partial')
            backend.strategy.send_email_validation(backend,
                                                   details['email'],
                                                   current_partial.token)
            backend.strategy.session_set('email_validation_address',
                                         details['email'])
            return backend.strategy.redirect(
                backend.strategy.setting('EMAIL_VALIDATION_URL')
            )



import json

from .oauth import OAuth2Test


class WunderlistOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.wunderlist.WunderlistOAuth2'
    user_data_url = 'https://a.wunderlist.com/api/v1/user'
    expected_username = '12345'
    access_token_body = json.dumps({
        'access_token': 'foobar-token',
        'token_type': 'foobar'})
    user_data_body = json.dumps({
        'created_at': '2015-01-21T00:56:51.442Z',
        'email': 'foo@bar.com',
        'id': 12345,
        'name': 'foobar',
        'revision': 1,
        'type': 'user',
        'updated_at': '2015-01-21T00:56:51.442Z'})

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json
import requests
import unittest2 as unittest

from httpretty import HTTPretty

from six.moves.urllib_parse import urlparse

from ...utils import parse_qs, module_member
from ...actions import do_auth, do_complete
from ..models import TestStorage, User, TestUserSocialAuth, TestNonce, \
                     TestAssociation
from ..strategy import TestStrategy


class BaseActionTest(unittest.TestCase):
    user_data_url = 'https://api.github.com/user'
    login_redirect_url = '/success'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'login': 'foobar',
        'id': 1,
        'avatar_url': 'https://github.com/images/error/foobar_happy.gif',
        'gravatar_id': 'somehexcode',
        'url': 'https://api.github.com/users/foobar',
        'name': 'monalisa foobar',
        'company': 'GitHub',
        'blog': 'https://github.com/blog',
        'location': 'San Francisco',
        'email': 'foo@bar.com',
        'hireable': False,
        'bio': 'There once was...',
        'public_repos': 2,
        'public_gists': 1,
        'followers': 20,
        'following': 0,
        'html_url': 'https://github.com/foobar',
        'created_at': '2008-01-14T04:33:35Z',
        'type': 'User',
        'total_private_repos': 100,
        'owned_private_repos': 100,
        'private_gists': 81,
        'disk_usage': 10000,
        'collaborators': 8,
        'plan': {
            'name': 'Medium',
            'space': 400,
            'collaborators': 10,
            'private_repos': 20
        }
    })

    def __init__(self, *args, **kwargs):
        self.strategy = None
        super(BaseActionTest, self).__init__(*args, **kwargs)

    def setUp(self):
        HTTPretty.enable()
        User.reset_cache()
        TestUserSocialAuth.reset_cache()
        TestNonce.reset_cache()
        TestAssociation.reset_cache()
        Backend = module_member('social_core.backends.github.GithubOAuth2')
        self.strategy = self.strategy or TestStrategy(TestStorage)
        self.backend = Backend(self.strategy, redirect_uri='/complete/github')
        self.user = None

    def tearDown(self):
        self.backend = None
        self.strategy = None
        self.user = None
        User.reset_cache()
        User.set_active(True)
        TestUserSocialAuth.reset_cache()
        TestNonce.reset_cache()
        TestAssociation.reset_cache()
        HTTPretty.disable()

    def do_login(self, after_complete_checks=True, user_data_body=None,
                 expected_username=None):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_KEY': 'a-key',
            'SOCIAL_AUTH_GITHUB_SECRET': 'a-secret-key',
            'SOCIAL_AUTH_LOGIN_REDIRECT_URL': self.login_redirect_url,
            'SOCIAL_AUTH_AUTHENTICATION_BACKENDS': (
                'social_core.backends.github.GithubOAuth2',
            )
        })
        start_url = do_auth(self.backend).url
        target_url = self.strategy.build_absolute_uri(
            '/complete/github/?code=foobar'
        )

        start_query = parse_qs(urlparse(start_url).query)
        location_url = target_url + ('&' if '?' in target_url else '?') + \
                       'state=' + start_query['state']
        location_query = parse_qs(urlparse(location_url).query)

        HTTPretty.register_uri(HTTPretty.GET, start_url, status=301,
                               location=location_url)
        HTTPretty.register_uri(HTTPretty.GET, location_url, status=200,
                               body='foobar')

        response = requests.get(start_url)
        self.assertEqual(response.url, location_url)
        self.assertEqual(response.text, 'foobar')

        HTTPretty.register_uri(HTTPretty.POST,
                               uri=self.backend.ACCESS_TOKEN_URL,
                               status=200,
                               body=self.access_token_body or '',
                               content_type='text/json')

        if self.user_data_url:
            user_data_body = user_data_body or self.user_data_body or ''
            HTTPretty.register_uri(HTTPretty.GET, self.user_data_url,
                                   body=user_data_body,
                                   content_type='text/json')
        self.strategy.set_request_data(location_query, self.backend)

        def _login(backend, user, social_user):
            backend.strategy.session_set('username', user.username)

        redirect = do_complete(self.backend, user=self.user, login=_login)

        if after_complete_checks:
            self.assertEqual(self.strategy.session_get('username'),
                             expected_username or self.expected_username)
            self.assertEqual(redirect.url, self.login_redirect_url)
        return redirect

    def do_login_with_partial_pipeline(self, before_complete=None):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_KEY': 'a-key',
            'SOCIAL_AUTH_GITHUB_SECRET': 'a-secret-key',
            'SOCIAL_AUTH_LOGIN_REDIRECT_URL': self.login_redirect_url,
            'SOCIAL_AUTH_AUTHENTICATION_BACKENDS': (
                'social_core.backends.github.GithubOAuth2',
            ),
            'SOCIAL_AUTH_PIPELINE': (
                'social_core.pipeline.social_auth.social_details',
                'social_core.pipeline.social_auth.social_uid',
                'social_core.pipeline.social_auth.auth_allowed',
                'social_core.tests.pipeline.ask_for_password',
                'social_core.pipeline.social_auth.social_user',
                'social_core.pipeline.user.get_username',
                'social_core.pipeline.user.create_user',
                'social_core.pipeline.social_auth.associate_user',
                'social_core.pipeline.social_auth.load_extra_data',
                'social_core.tests.pipeline.set_password',
                'social_core.pipeline.user.user_details'
            )
        })
        start_url = do_auth(self.backend).url
        target_url = self.strategy.build_absolute_uri(
            '/complete/github/?code=foobar'
        )

        start_query = parse_qs(urlparse(start_url).query)
        location_url = target_url + ('&' if '?' in target_url else '?') + \
                       'state=' + start_query['state']
        location_query = parse_qs(urlparse(location_url).query)

        HTTPretty.register_uri(HTTPretty.GET, start_url, status=301,
                               location=location_url)
        HTTPretty.register_uri(HTTPretty.GET, location_url, status=200,
                               body='foobar')

        response = requests.get(start_url)
        self.assertEqual(response.url, location_url)
        self.assertEqual(response.text, 'foobar')

        HTTPretty.register_uri(HTTPretty.GET,
                               uri=self.backend.ACCESS_TOKEN_URL,
                               status=200,
                               body=self.access_token_body or '',
                               content_type='text/json')

        if self.user_data_url:
            HTTPretty.register_uri(HTTPretty.GET, self.user_data_url,
                                   body=self.user_data_body or '',
                                   content_type='text/json')
        self.strategy.set_request_data(location_query, self.backend)

        def _login(backend, user, social_user):
            backend.strategy.session_set('username', user.username)

        redirect = do_complete(self.backend, user=self.user, login=_login)
        url = self.strategy.build_absolute_uri('/password')
        self.assertEqual(redirect.url, url)
        HTTPretty.register_uri(HTTPretty.GET, redirect.url, status=200,
                               body='foobar')
        HTTPretty.register_uri(HTTPretty.POST, redirect.url, status=200)

        password = 'foobar'
        requests.get(url)
        requests.post(url, data={'password': password})
        data = parse_qs(HTTPretty.last_request.body)
        self.assertEqual(data['password'], password)
        self.strategy.session_set('password', data['password'])

        if before_complete:
            before_complete()
        redirect = do_complete(self.backend, user=self.user, login=_login)
        self.assertEqual(self.strategy.session_get('username'),
                         self.expected_username)
        self.assertEqual(redirect.url, self.login_redirect_url)



import json

from .oauth import OAuth2Test


class YandexOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.yandex.YandexOAuth2'
    user_data_url = 'https://login.yandex.ru/info'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'display_name': 'foobar',
        'real_name': 'Foo Bar',
        'sex': None,
        'id': '101010101',
        'default_email': 'foobar@yandex.com',
        'emails': ['foobar@yandex.com']
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



from pprint import pprint


def debug(response, details, *args, **kwargs):
    print('=' * 80)
    pprint(response)
    print('=' * 80)
    pprint(details)
    print('=' * 80)
    pprint(args)
    print('=' * 80)
    pprint(kwargs)
    print('=' * 80)



"""
Lyft OAuth2 backend. Read more about the
    API at https://developer.lyft.com/docs
"""
from .oauth import BaseOAuth2


class LyftOAuth2(BaseOAuth2):
    name = 'lyft'
    ID_KEY = 'id'
    SCOPE_SEPARATOR = ' '
    AUTHORIZATION_URL = 'https://api.lyft.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://api.lyft.com/oauth/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REFRESH_TOKEN_URL = 'https://api.lyft.com/oauth/token'
    USER_DATA_URL = 'https://api.lyft.com/v1/profile'
    DEFAULT_SCOPE = ['public', 'profile', 'rides.read', 'rides.request']
    RESPONSE_TYPE = 'code'
    STATE_PARAMETER = 'asdf'
    EXTRA_DATA = [
        ('id', 'id'),
        ('username', 'username'),
        ('access_token', 'access_token'),
        ('refresh_token', 'refresh_token'),
        ('token_type', 'token_type'),
        ('expires_in', 'expires_in'),
        ('scope', 'scope'),
    ]

    def get_user_details(self, response):
        """Return user details from Lyft account"""
        return {
            'id': response['id'],
            'username': response['id']
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        response = kwargs.pop('response')
        return self.get_json(self.USER_DATA_URL, headers={
          'Authorization': 'Bearer {0}'.format(access_token)
        })

    def auth_complete_params(self, state=None):
        client_id, client_secret = self.get_key_and_secret()
        return {
            'grant_type': 'authorization_code',
            'code': self.data['code']
        }

    def auth_complete_credentials(self):
        return self.get_key_and_secret()

    def refresh_token_params(self, refresh_token, *args, **kwargs):
        return {
            'refresh_token': refresh_token,
            'grant_type': 'refresh_token'
        }



"""
Quizlet OAuth2 Sign-in backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/quizlet.html
"""
from .oauth import BaseOAuth2


class QuizletOAuth2(BaseOAuth2):
    """Quizlet OAuth2"""
    name = 'quizlet'
    ID_KEY = 'user_id'
    API_URL = 'https://api.quizlet.com/2.0/'
    AUTHORIZATION_URL = 'https://quizlet.com/authorize'
    ACCESS_TOKEN_URL = 'https://api.quizlet.com/oauth/token'
    ACCESS_TOKEN_METHOD = 'POST'
    SCOPE_SEPARATOR = ' '
    DEFAULT_SCOPE = ['read']

    def get_user_details(self, response):
        """Return user details from Quizlet account"""
        return {
            'username': response.get('user_id')
        }



"""
Reddit OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/reddit.html
"""
import base64

from .oauth import BaseOAuth2


class RedditOAuth2(BaseOAuth2):
    """Reddit OAuth2 authentication backend"""
    name = 'reddit'
    AUTHORIZATION_URL = 'https://ssl.reddit.com/api/v1/authorize'
    ACCESS_TOKEN_URL = 'https://ssl.reddit.com/api/v1/access_token'
    ACCESS_TOKEN_METHOD = 'POST'
    REFRESH_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    SCOPE_SEPARATOR = ','
    DEFAULT_SCOPE = ['identity']
    SEND_USER_AGENT = True
    EXTRA_DATA = [
        ('id', 'id'),
        ('name', 'username'),
        ('link_karma', 'link_karma'),
        ('comment_karma', 'comment_karma'),
        ('refresh_token', 'refresh_token'),
        ('expires_in', 'expires')
    ]

    def get_user_details(self, response):
        """Return user details from Reddit account"""
        return {'username': response.get('name'),
                'email': '', 'fullname': '',
                'first_name': '', 'last_name': ''}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            'https://oauth.reddit.com/api/v1/me.json',
            headers={'Authorization': 'bearer ' + access_token}
        )

    def auth_headers(self):
        return {
            'Authorization': b'Basic ' + base64.urlsafe_b64encode(
                '{0}:{1}'.format(*self.get_key_and_secret()).encode()
            )
        }

    def refresh_token_params(self, token, redirect_uri=None, *args, **kwargs):
        params = super(RedditOAuth2, self).refresh_token_params(token)
        params['redirect_uri'] = self.redirect_uri or redirect_uri
        return params



import json

from .oauth import OAuth2Test


class DripOAuthTest(OAuth2Test):
    backend_path = 'social_core.backends.drip.DripOAuth'
    user_data_url = 'https://api.getdrip.com/v2/user'
    expected_username = 'other@example.com'
    access_token_body = json.dumps({
        'access_token': '822bbf7cd12243df',
        'token_type': 'bearer',
        'scope': 'public'
    })

    user_data_body = json.dumps({
        'users': [
            {
                'email': 'other@example.com',
                'name': None
            }
        ]
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



from functools import wraps

from .utils import partial_prepare


def partial(func):
    """Wraps func to behave like a partial pipeline step, any output
    that's not None or {} will be considered a response object and
    will be returned to user.

    The pipeline function will receive a current_partial object, it
    contains the partial pipeline data and a token that is used to
    identify it when it's continued, this is useful to build links
    with the token.

    The default value for this parameter is partial_token, but can be
    overridden by SOCIAL_AUTH_PARTIAL_PIPELINE_TOKEN_NAME setting.

    The token is also stored in the session under the
    partial_pipeline_token key.
    """
    @wraps(func)
    def wrapper(strategy, backend, pipeline_index, *args, **kwargs):
        current_partial = partial_prepare(strategy, backend, pipeline_index,
                                          *args, **kwargs)

        out = func(strategy=strategy,
                   backend=backend,
                   pipeline_index=pipeline_index,
                   current_partial=current_partial,
                   *args, **kwargs) or {}

        if not isinstance(out, dict):
            strategy.storage.partial.store(current_partial)
            strategy.session_set('partial_pipeline_token',
                                 current_partial.token)
        return out
    return wrapper



import json

from six.moves.urllib_parse import urlencode

from .oauth import OAuth2Test


class DatagouvfrOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.udata.DatagouvfrOAuth2'
    user_data_url = 'https://www.data.gouv.fr/api/1/me/'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer',
        'first_name': 'foobar',
        'email': 'foobar@example.com'
    })
    request_token_body = urlencode({
        'oauth_token_secret': 'foobar-secret',
        'oauth_token': 'foobar',
        'oauth_callback_confirmed': 'true'
    })
    user_data_body = json.dumps({})

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



# -*- coding: utf-8 -*-
from calendar import timegm

import os
import sys
import json
import datetime
import unittest2

try:
    from jwkest.jwk import RSAKey, KEYS
    from jwkest.jws import JWS
    from jwkest.jwt import b64encode_item
    NO_JWKEST = False
except ImportError:
    NO_JWKEST = True

from httpretty import HTTPretty

sys.path.insert(0, '..')

from ...exceptions import AuthTokenError


class OpenIdConnectTestMixin(object):
    """
    Mixin to test OpenID Connect consumers. Inheriting classes should also
    inherit OAuth2Test.
    """
    client_key = 'a-key'
    client_secret = 'a-secret-key'
    issuer = None  # id_token issuer
    openid_config_body = None
    key = None

    def setUp(self):
        super(OpenIdConnectTestMixin, self).setUp()
        test_root = os.path.dirname(os.path.dirname(__file__))
        self.key = RSAKey(kid='testkey').load(os.path.join(test_root, 'testkey.pem'))
        HTTPretty.register_uri(HTTPretty.GET,
          self.backend.OIDC_ENDPOINT + '/.well-known/openid-configuration',
          status=200,
          body=self.openid_config_body
        )
        oidc_config = json.loads(self.openid_config_body)

        def jwks(_request, _uri, headers):
            ks = KEYS()
            ks.add(self.key.serialize())
            return 200, headers, ks.dump_jwks()

        HTTPretty.register_uri(HTTPretty.GET,
                               oidc_config.get('jwks_uri'),
                               status=200,
                               body=jwks)

    def extra_settings(self):
        settings = super(OpenIdConnectTestMixin, self).extra_settings()
        settings.update({
            'SOCIAL_AUTH_{0}_KEY'.format(self.name): self.client_key,
            'SOCIAL_AUTH_{0}_SECRET'.format(self.name): self.client_secret,
            'SOCIAL_AUTH_{0}_ID_TOKEN_DECRYPTION_KEY'.format(self.name):
                self.client_secret
        })
        return settings

    def access_token_body(self, request, _url, headers):
        """
        Get the nonce from the request parameters, add it to the id_token, and
        return the complete response.
        """
        nonce = self.backend.data['nonce'].encode('utf-8')
        body = self.prepare_access_token_body(nonce=nonce)
        return 200, headers, body

    def get_id_token(self, client_key=None, expiration_datetime=None,
                     issue_datetime=None, nonce=None, issuer=None):
        """
        Return the id_token to be added to the access token body.
        """
        return {
            'iss': issuer,
            'nonce': nonce,
            'aud': client_key,
            'azp': client_key,
            'exp': expiration_datetime,
            'iat': issue_datetime,
            'sub': '1234'
        }

    def prepare_access_token_body(self, client_key=None, tamper_message=False,
                                  expiration_datetime=None,
                                  issue_datetime=None, nonce=None,
                                  issuer=None):
        """
        Prepares a provider access token response. Arguments:

        client_id       -- (str) OAuth ID for the client that requested
                                 authentication.
        expiration_time -- (datetime) Date and time after which the response
                                      should be considered invalid.
        """

        body = {'access_token': 'foobar', 'token_type': 'bearer'}
        client_key = client_key or self.client_key
        now = datetime.datetime.utcnow()
        expiration_datetime = expiration_datetime or \
                              (now + datetime.timedelta(seconds=30))
        issue_datetime = issue_datetime or now
        nonce = nonce or 'a-nonce'
        issuer = issuer or self.issuer
        id_token = self.get_id_token(
            client_key, timegm(expiration_datetime.utctimetuple()),
            timegm(issue_datetime.utctimetuple()), nonce, issuer)

        body['id_token'] = JWS(id_token, jwk=self.key, alg='RS256').sign_compact()
        if tamper_message:
            header, msg, sig = body['id_token'].split('.')
            id_token['sub'] = '1235'
            msg = b64encode_item(id_token).decode('utf-8')
            body['id_token'] = '.'.join([header, msg, sig])

        return json.dumps(body)

    def authtoken_raised(self, expected_message, **access_token_kwargs):
        self.access_token_body = self.prepare_access_token_body(
            **access_token_kwargs
        )
        with self.assertRaisesRegexp(AuthTokenError, expected_message):
            self.do_login()

    @unittest2.skipIf(NO_JWKEST, 'No Jwkest installed')
    def test_invalid_signature(self):
        self.authtoken_raised(
            'Token error: Signature verification failed',
            tamper_message=True
        )

    @unittest2.skipIf(NO_JWKEST, 'No Jwkest installed')
    def test_expired_signature(self):
        expiration_datetime = datetime.datetime.utcnow() - \
                              datetime.timedelta(seconds=30)
        self.authtoken_raised('Token error: Signature has expired',
                              expiration_datetime=expiration_datetime)

    @unittest2.skipIf(NO_JWKEST, 'No Jwkest installed')
    def test_invalid_issuer(self):
        self.authtoken_raised('Token error: Invalid issuer',
                              issuer='someone-else')

    @unittest2.skipIf(NO_JWKEST, 'No Jwkest installed')
    def test_invalid_audience(self):
        self.authtoken_raised('Token error: Invalid audience',
                              client_key='someone-else')

    @unittest2.skipIf(NO_JWKEST, 'No Jwkest installed')
    def test_invalid_issue_time(self):
        expiration_datetime = datetime.datetime.utcnow() - \
                              datetime.timedelta(hours=1)
        self.authtoken_raised('Token error: Incorrect id_token: iat',
                              issue_datetime=expiration_datetime)

    @unittest2.skipIf(NO_JWKEST, 'No Jwkest installed')
    def test_invalid_nonce(self):
        self.authtoken_raised(
            'Token error: Incorrect id_token: nonce',
            nonce='something-wrong'
        )



import json

from six.moves.urllib_parse import urlencode

from .oauth import OAuth1Test


class TripitOAuth1Test(OAuth1Test):
    backend_path = 'social_core.backends.tripit.TripItOAuth'
    user_data_url = 'https://api.tripit.com/v1/get/profile'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    request_token_body = urlencode({
        'oauth_token_secret': 'foobar-secret',
        'oauth_token': 'foobar',
        'oauth_callback_confirmed': 'true'
    })
    user_data_content_type = 'text/xml'
    user_data_body = \
        '<Response>' \
            '<timestamp>1363590451</timestamp>' \
            '<num_bytes>1040</num_bytes>' \
            '<Profile ref="ignore-me">' \
                '<ProfileEmailAddresses>' \
                    '<ProfileEmailAddress>' \
                        '<address>foobar@gmail.com</address>' \
                        '<is_auto_import>false</is_auto_import>' \
                        '<is_confirmed>true</is_confirmed>' \
                        '<is_primary>true</is_primary>' \
                        '<is_auto_inbox_eligible>' \
                            'true' \
                        '</is_auto_inbox_eligible>' \
                    '</ProfileEmailAddress>' \
                '</ProfileEmailAddresses>' \
                '<is_client>true</is_client>' \
                '<is_pro>false</is_pro>' \
                '<screen_name>foobar</screen_name>' \
                '<public_display_name>Foo Bar</public_display_name>' \
                '<profile_url>people/foobar</profile_url>' \
                '<home_city>Foo, Barland</home_city>' \
                '<activity_feed_url>' \
                    'https://www.tripit.com/feed/activities/private/' \
                    'ignore-this/activities.atom' \
                '</activity_feed_url>' \
                '<alerts_feed_url>' \
                    'https://www.tripit.com/feed/alerts/private/' \
                    'ignore-this/alerts.atom' \
                '</alerts_feed_url>' \
                '<ical_url>' \
                    'webcal://www.tripit.com/feed/ical/private/' \
                    'ignore-this/tripit.ics' \
                '</ical_url>' \
            '</Profile>' \
        '</Response>'

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()


class TripitOAuth1UsernameAlternativesTest(TripitOAuth1Test):
    user_data_body = \
        '<Response>' \
            '<timestamp>1363590451</timestamp>' \
            '<num_bytes>1040</num_bytes>' \
            '<Profile ref="ignore-me">' \
                '<ProfileEmailAddresses>' \
                    '<ProfileEmailAddress>' \
                        '<address>foobar@gmail.com</address>' \
                        '<is_auto_import>false</is_auto_import>' \
                        '<is_confirmed>true</is_confirmed>' \
                        '<is_primary>true</is_primary>' \
                        '<is_auto_inbox_eligible>' \
                            'true' \
                        '</is_auto_inbox_eligible>' \
                    '</ProfileEmailAddress>' \
                '</ProfileEmailAddresses>' \
                '<is_client>true</is_client>' \
                '<is_pro>false</is_pro>' \
                '<screen_name>foobar</screen_name>' \
                '<public_display_name>Foobar</public_display_name>' \
                '<profile_url>people/foobar</profile_url>' \
                '<home_city>Foo, Barland</home_city>' \
                '<activity_feed_url>' \
                    'https://www.tripit.com/feed/activities/private/' \
                    'ignore-this/activities.atom' \
                '</activity_feed_url>' \
                '<alerts_feed_url>' \
                    'https://www.tripit.com/feed/alerts/private/' \
                    'ignore-this/alerts.atom' \
                '</alerts_feed_url>' \
                '<ical_url>' \
                    'webcal://www.tripit.com/feed/ical/private/' \
                    'ignore-this/tripit.ics' \
                '</ical_url>' \
            '</Profile>' \
        '</Response>'



"""
NationBuilder OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/nationbuilder.html
"""
from .oauth import BaseOAuth2


class NationBuilderOAuth2(BaseOAuth2):
    """NationBuilder OAuth2 authentication backend"""
    name = 'nationbuilder'
    AUTHORIZATION_URL = 'https://{slug}.nationbuilder.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://{slug}.nationbuilder.com/oauth/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    SCOPE_SEPARATOR = ','
    EXTRA_DATA = [
        ('id', 'id'),
        ('expires', 'expires')
    ]

    def authorization_url(self):
        return self.AUTHORIZATION_URL.format(slug=self.slug)

    def access_token_url(self):
        return self.ACCESS_TOKEN_URL.format(slug=self.slug)

    @property
    def slug(self):
        return self.setting('SLUG')

    def get_user_details(self, response):
        """Return user details from Github account"""
        email = response.get('email') or ''
        username = email.split('@')[0] if email else ''
        return {'username': username,
                'email': email,
                'fullname': response.get('full_name') or '',
                'first_name': response.get('first_name') or '',
                'last_name': response.get('last_name') or ''}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        url = 'https://{slug}.nationbuilder.com/api/v1/people/me'.format(
            slug=self.slug
        )
        return self.get_json(url, params={
            'access_token': access_token
        })['person']



from urllib import urlencode

import six

from requests_oauthlib import OAuth1

from .oauth import BaseOAuth2


class NKOAuth2(BaseOAuth2):
    """NK OAuth authentication backend"""
    name = 'nk'
    AUTHORIZATION_URL = 'https://nk.pl/oauth2/login'
    ACCESS_TOKEN_URL = 'https://nk.pl/oauth2/token'
    SCOPE_SEPARATOR = ','
    ACCESS_TOKEN_METHOD = 'POST'
    SIGNATURE_TYPE_AUTH_HEADER = 'AUTH_HEADER'
    EXTRA_DATA = [
        ('id', 'id'),
    ]

    def get_user_details(self, response):
        """Return user details from NK account"""
        entry = response['entry']
        return {
            'username': entry.get('displayName'),
            'email': entry['emails'][0]['value'],
            'first_name': entry.get('displayName').split(' ')[0],
            'id': entry.get('id')
        }

    def auth_complete_params(self, state=None):
        client_id, client_secret = self.get_key_and_secret()
        return {
            'grant_type': 'authorization_code',  # request auth code
            'code': self.data.get('code', ''),  # server response code
            'client_id': client_id,
            'client_secret': client_secret,
            'redirect_uri': self.get_redirect_uri(state),
            'scope': self.get_scope_argument()
        }

    def get_user_id(self, details, response):
        """Return a unique ID for the current user, by default from server
        response."""
        return details.get(self.ID_KEY)

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        url = 'http://opensocial.nk-net.pl/v09/social/rest/people/@me?' + \
              urlencode({
                  'nk_token': access_token,
                  'fields': 'name,surname,avatar,localization,age,' +
                            'gender,emails,birthdate'
              })
        return self.get_json(
            url,
            auth=self.oauth_auth(access_token)
        )

    def oauth_auth(self, token=None, oauth_verifier=None,
                   signature_type=SIGNATURE_TYPE_AUTH_HEADER):
        key, secret = self.get_key_and_secret()
        oauth_verifier = oauth_verifier or self.data.get('oauth_verifier')
        token = token or {}
        # decoding='utf-8' produces errors with python-requests on Python3
        # since the final URL will be of type bytes
        decoding = None if six.PY3 else 'utf-8'
        state = self.get_or_create_state()
        return OAuth1(key, secret,
                      resource_owner_key=None,
                      resource_owner_secret=None,
                      callback_uri=self.get_redirect_uri(state),
                      verifier=oauth_verifier,
                      signature_type=signature_type,
                      decoding=decoding)



"""
Flickr OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/flickr.html
"""
from .oauth import BaseOAuth1


class FlickrOAuth(BaseOAuth1):
    """Flickr OAuth authentication backend"""
    name = 'flickr'
    AUTHORIZATION_URL = 'https://www.flickr.com/services/oauth/authorize'
    REQUEST_TOKEN_URL = 'https://www.flickr.com/services/oauth/request_token'
    ACCESS_TOKEN_URL = 'https://www.flickr.com/services/oauth/access_token'
    EXTRA_DATA = [
        ('id', 'id'),
        ('username', 'username'),
        ('expires', 'expires')
    ]

    def get_user_details(self, response):
        """Return user details from Flickr account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get('fullname')
        )
        return {'username': response.get('username') or response.get('id'),
                'email': '',
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return {
            'id': access_token['user_nsid'],
            'username': access_token['username'],
            'fullname': access_token.get('fullname', ''),
        }

    def auth_extra_arguments(self):
        params = super(FlickrOAuth, self).auth_extra_arguments() or {}
        if 'perms' not in params:
            params['perms'] = 'read'
        return params



"""
GitLab OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/gitlab.html
"""
from requests import HTTPError

from six.moves.urllib.parse import urljoin

from .oauth import BaseOAuth2
from ..exceptions import AuthFailed


class GitLabOAuth2(BaseOAuth2):
    """GitLab OAuth authentication backend"""
    name = 'gitlab'
    API_URL = 'https://gitlab.com'
    AUTHORIZATION_URL = 'https://gitlab.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://gitlab.com/oauth/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    DEFAULT_SCOPE = ['read_user']
    EXTRA_DATA = [
        ('id', 'id'),
        ('expires_in', 'expires'),
        ('refresh_token', 'refresh_token')
    ]

    def api_url(self, path):
        api_url = self.setting('API_URL') or self.API_URL
        return '{0}{1}'.format(api_url.rstrip('/'), path)

    def authorization_url(self):
        return self.api_url('/oauth/authorize')

    def access_token_url(self):
        return self.api_url('/oauth/token')

    def get_user_details(self, response):
        """Return user details from GitLab account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get('name')
        )
        return {'username': response.get('username'),
                'email': response.get('email') or '',
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(self.api_url('/api/v3/user'), params={
            'access_token': access_token
        })



import json

from six.moves.urllib_parse import urlencode


from .oauth import OAuth1Test, OAuth2Test


class BaseLinkedinTest(object):
    user_data_url = 'https://api.linkedin.com/v1/people/~:' \
                        '(first-name,id,last-name)'
    expected_username = 'FooBar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'lastName': 'Bar',
        'id': '1010101010',
        'firstName': 'Foo'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()


class LinkedinOAuth1Test(BaseLinkedinTest, OAuth1Test):
    backend_path = 'social_core.backends.linkedin.LinkedinOAuth'
    request_token_body = urlencode({
        'oauth_token_secret': 'foobar-secret',
        'oauth_token': 'foobar',
        'oauth_callback_confirmed': 'true'
    })


class LinkedinOAuth2Test(BaseLinkedinTest, OAuth2Test):
    backend_path = 'social_core.backends.linkedin.LinkedinOAuth2'



# -*- coding: utf-8 -*-
# author:duoduo3369@gmail.com  https://github.com/duoduo369
"""
Weixin OAuth2 backend
"""
import urllib
from requests import HTTPError

from .oauth import BaseOAuth2
from ..exceptions import AuthCanceled, AuthUnknownError


class WeixinOAuth2(BaseOAuth2):
    """Weixin OAuth authentication backend"""
    name = 'weixin'
    ID_KEY = 'openid'
    AUTHORIZATION_URL = 'https://open.weixin.qq.com/connect/qrconnect'
    ACCESS_TOKEN_URL = 'https://api.weixin.qq.com/sns/oauth2/access_token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ('nickname', 'username'),
        ('headimgurl', 'profile_image_url'),
    ]

    def get_user_details(self, response):
        """Return user details from Weixin. API URL is:
        https://api.weixin.qq.com/sns/userinfo
        """
        if self.setting('DOMAIN_AS_USERNAME'):
            username = response.get('domain', '')
        else:
            username = response.get('nickname', '')
        return {
            'username': username,
            'profile_image_url': response.get('headimgurl', '')
        }

    def user_data(self, access_token, *args, **kwargs):
        data = self.get_json('https://api.weixin.qq.com/sns/userinfo', params={
            'access_token': access_token,
            'openid': kwargs['response']['openid']
        })
        nickname = data.get('nickname')
        if nickname:
            # weixin api has some encode bug, here need handle
            data['nickname'] = nickname.encode(
                'raw_unicode_escape'
            ).decode('utf-8')
        return data

    def auth_params(self, state=None):
        appid, secret = self.get_key_and_secret()
        params = {
            'appid': appid,
            'redirect_uri': self.get_redirect_uri(state)
        }
        if self.STATE_PARAMETER and state:
            params['state'] = state
        if self.RESPONSE_TYPE:
            params['response_type'] = self.RESPONSE_TYPE
        return params

    def auth_complete_params(self, state=None):
        appid, secret = self.get_key_and_secret()
        return {
            'grant_type': 'authorization_code',  # request auth code
            'code': self.data.get('code', ''),  # server response code
            'appid': appid,
            'secret': secret,
            'redirect_uri': self.get_redirect_uri(state)
        }

    def refresh_token_params(self, token, *args, **kwargs):
        appid, secret = self.get_key_and_secret()
        return {
            'refresh_token': token,
            'grant_type': 'refresh_token',
            'appid': appid,
            'secret': secret
        }

    def auth_complete(self, *args, **kwargs):
        """Completes loging process, must return user instance"""
        self.process_error(self.data)
        try:
            response = self.request_access_token(
                self.ACCESS_TOKEN_URL,
                data=self.auth_complete_params(self.validate_state()),
                headers=self.auth_headers(),
                method=self.ACCESS_TOKEN_METHOD
            )
        except HTTPError as err:
            if err.response.status_code == 400:
                raise AuthCanceled(self, response=err.response)
            else:
                raise
        except KeyError:
            raise AuthUnknownError(self)
        if 'errcode' in response:
            raise AuthCanceled(self)
        self.process_error(response)
        return self.do_auth(response['access_token'], response=response,
                            *args, **kwargs)


class WeixinOAuth2APP(WeixinOAuth2):
    """
    Weixin OAuth authentication backend

    Can't use in web, only in weixin app
    """
    name = 'weixinapp'
    ID_KEY = 'openid'
    AUTHORIZATION_URL = 'https://open.weixin.qq.com/connect/oauth2/authorize'
    ACCESS_TOKEN_URL = 'https://api.weixin.qq.com/sns/oauth2/access_token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False

    def auth_url(self):
        if self.STATE_PARAMETER or self.REDIRECT_STATE:
            # Store state in session for further request validation. The state
            # value is passed as state parameter (as specified in OAuth2 spec),
            # but also added to redirect, that way we can still verify the
            # request if the provider doesn't implement the state parameter.
            # Reuse token if any.
            name = self.name + '_state'
            state = self.strategy.session_get(name)
            if state is None:
                state = self.state_token()
                self.strategy.session_set(name, state)
        else:
            state = None

        params = self.auth_params(state)
        params.update(self.get_scope_argument())
        params.update(self.auth_extra_arguments())
        params = urllib.urlencode(sorted(params.items()))
        return '{}#wechat_redirect'.format(
            self.AUTHORIZATION_URL + '?' + params
        )

    def auth_complete_params(self, state=None):
        appid, secret = self.get_key_and_secret()
        return {
            'grant_type': 'authorization_code',  # request auth code
            'code': self.data.get('code', ''),  # server response code
            'appid': appid,
            'secret': secret,
        }

    def validate_state(self):
        return None

    def auth_complete(self, *args, **kwargs):
        """Completes loging process, must return user instance"""
        self.process_error(self.data)
        try:
            response = self.request_access_token(
                self.ACCESS_TOKEN_URL,
                data=self.auth_complete_params(self.validate_state()),
                headers=self.auth_headers(),
                method=self.ACCESS_TOKEN_METHOD
            )
        except HTTPError as err:
            if err.response.status_code == 400:
                raise AuthCanceled(self)
            else:
                raise
        except KeyError:
            raise AuthUnknownError(self)

        if 'errcode' in response:
            raise AuthCanceled(self)
        self.process_error(response)
        return self.do_auth(response['access_token'], response=response,
                            *args, **kwargs)



import json

from .oauth import OAuth2Test


class MapMyFitnessOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.mapmyfitness.MapMyFitnessOAuth2'
    user_data_url = 'https://oauth2-api.mapmyapi.com/v7.0/user/self/'
    expected_username = 'FredFlinstone'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'Bearer',
        'expires_in': 4000000,
        'refresh_token': 'bambaz',
        'scope': 'read'
    })
    user_data_body = json.dumps({
        'last_name': 'Flinstone',
        'weight': 91.17206637,
        'communication': {
            'promotions': True,
            'newsletter': True,
            'system_messages': True
        },
        'height': 1.778,
        'token_type': 'Bearer',
        'id': 112233,
        'date_joined': '2011-08-26T06:06:19+00:00',
        'first_name': 'Fred',
        'display_name': 'Fred Flinstone',
        'display_measurement_system': 'imperial',
        'expires_in': 4000000,
        '_links': {
            'stats': [
                {
                    'href': '/v7.0/user_stats/112233/?'
                            'aggregate_by_period=month',
                    'id': '112233',
                    'name': 'month'
                },
                {
                    'href': '/v7.0/user_stats/112233/?'
                            'aggregate_by_period=year',
                    'id': '112233',
                    'name': 'year'
                },
                {
                    'href': '/v7.0/user_stats/112233/?aggregate_by_period=day',
                    'id': '112233',
                    'name': 'day'
                },
                {
                    'href': '/v7.0/user_stats/112233/?'
                            'aggregate_by_period=week',
                    'id': '112233',
                    'name': 'week'
                },
                {
                    'href': '/v7.0/user_stats/112233/?'
                            'aggregate_by_period=lifetime',
                    'id': '112233',
                    'name': 'lifetime'
                }
            ],
            'friendships': [
                {
                    'href': '/v7.0/friendship/?from_user=112233'
                }
            ],
            'privacy': [
                {
                    'href': '/v7.0/privacy_option/3/',
                    'id': '3',
                    'name': 'profile'
                },
                {
                    'href': '/v7.0/privacy_option/3/',
                    'id': '3',
                    'name': 'workout'
                },
                {
                    'href': '/v7.0/privacy_option/3/',
                    'id': '3',
                    'name': 'activity_feed'
                },
                {
                    'href': '/v7.0/privacy_option/1/',
                    'id': '1',
                    'name': 'food_log'
                },
                {
                    'href': '/v7.0/privacy_option/3/',
                    'id': '3',
                    'name': 'email_search'
                },
                {
                    'href': '/v7.0/privacy_option/3/',
                    'id': '3',
                    'name': 'route'
                }
            ],
            'image': [
                {
                    'href': '/v7.0/user_profile_photo/112233/',
                    'id': '112233',
                    'name': 'user_profile_photo'
                }
            ],
            'documentation': [
                {
                    'href': 'https://www.mapmyapi.com/docs/User'
                }
            ],
            'workouts': [
                {
                    'href': '/v7.0/workout/?user=112233&'
                            'order_by=-start_datetime'
                }
            ],
            'deactivation': [
                {
                    'href': '/v7.0/user_deactivation/'
                }
            ],
            'self': [
                {
                    'href': '/v7.0/user/112233/',
                    'id': '112233'
                }
            ]
        },
        'location': {
            'country': 'US',
            'region': 'NC',
            'locality': 'Bedrock',
            'address': '150 Dinosaur Ln'
        },
        'last_login': '2014-02-23T22:36:52+00:00',
        'email': 'fredflinstone@gmail.com',
        'username': 'FredFlinstone',
        'sharing': {
            'twitter': False,
            'facebook': False
        },
        'scope': 'read',
        'refresh_token': 'bambaz',
        'last_initial': 'S.',
        'access_token': 'foobar',
        'gender': 'M',
        'time_zone': 'America/Denver',
        'birthdate': '1983-04-15'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import unittest2 as unittest

from ..exceptions import SocialAuthBaseException, WrongBackend, \
                         AuthFailed, AuthTokenError, \
                         AuthMissingParameter, AuthStateMissing, \
                         NotAllowedToDisconnect, AuthException, \
                         AuthCanceled, AuthUnknownError, \
                         AuthStateForbidden, AuthAlreadyAssociated, \
                         AuthTokenRevoked, AuthForbidden, \
                         AuthUnreachableProvider, InvalidEmail, \
                         MissingBackend


class BaseExceptionTestCase(unittest.TestCase):
    exception = None
    expected_message = ''

    def test_exception_message(self):
        if self.exception is None and self.expected_message == '':
            return
        try:
            raise self.exception
        except SocialAuthBaseException as err:
            self.assertEqual(str(err), self.expected_message)


class WrongBackendTest(BaseExceptionTestCase):
    exception = WrongBackend('foobar')
    expected_message = 'Incorrect authentication service "foobar"'


class AuthFailedTest(BaseExceptionTestCase):
    exception = AuthFailed('foobar', 'wrong_user')
    expected_message = 'Authentication failed: wrong_user'


class AuthFailedDeniedTest(BaseExceptionTestCase):
    exception = AuthFailed('foobar', 'access_denied')
    expected_message = 'Authentication process was canceled'


class AuthTokenErrorTest(BaseExceptionTestCase):
    exception = AuthTokenError('foobar', 'Incorrect tokens')
    expected_message = 'Token error: Incorrect tokens'


class AuthMissingParameterTest(BaseExceptionTestCase):
    exception = AuthMissingParameter('foobar', 'username')
    expected_message = 'Missing needed parameter username'


class AuthStateMissingTest(BaseExceptionTestCase):
    exception = AuthStateMissing('foobar')
    expected_message = 'Session value state missing.'


class NotAllowedToDisconnectTest(BaseExceptionTestCase):
    exception = NotAllowedToDisconnect()
    expected_message = ''


class AuthExceptionTest(BaseExceptionTestCase):
    exception = AuthException('foobar', 'message')
    expected_message = 'message'


class AuthCanceledTest(BaseExceptionTestCase):
    exception = AuthCanceled('foobar')
    expected_message = 'Authentication process canceled'


class AuthCanceledWithExtraMessageTest(BaseExceptionTestCase):
    exception = AuthCanceled('foobar', 'error_message')
    expected_message = 'Authentication process canceled: error_message'


class AuthUnknownErrorTest(BaseExceptionTestCase):
    exception = AuthUnknownError('foobar', 'some error')
    expected_message = 'An unknown error happened while ' \
                       'authenticating some error'


class AuthStateForbiddenTest(BaseExceptionTestCase):
    exception = AuthStateForbidden('foobar')
    expected_message = 'Wrong state parameter given.'


class AuthAlreadyAssociatedTest(BaseExceptionTestCase):
    exception = AuthAlreadyAssociated('foobar')
    expected_message = ''


class AuthTokenRevokedTest(BaseExceptionTestCase):
    exception = AuthTokenRevoked('foobar')
    expected_message = 'User revoke access to the token'


class AuthForbiddenTest(BaseExceptionTestCase):
    exception = AuthForbidden('foobar')
    expected_message = 'Your credentials aren\'t allowed'


class AuthUnreachableProviderTest(BaseExceptionTestCase):
    exception = AuthUnreachableProvider('foobar')
    expected_message = 'The authentication provider could not be reached'


class InvalidEmailTest(BaseExceptionTestCase):
    exception = InvalidEmail('foobar')
    expected_message = 'Email couldn\'t be validated'


class MissingBackendTest(BaseExceptionTestCase):
    exception = MissingBackend('backend')
    expected_message = 'Missing backend "backend" entry'
